import ujson as json
import datetime
from dataclasses import dataclass, asdict, is_dataclass, fields
from collections import abc

from .util import JsonEncoder
from yankee.util import is_valid

def to_dict(obj, item_class=dict, collection_class=list, date_style="python"):
    if isinstance(obj, abc.Mapping):
        return item_class((k, to_dict(v, item_class, collection_class, date_style)) for k, v in obj.items())
    elif is_dataclass(obj):
        return to_dict(item_class(obj), item_class, collection_class, date_style)
    elif isinstance(obj, abc.Iterable) and not isinstance(obj, (str, bytes)):
        return collection_class(to_dict(i, item_class, collection_class, date_style) for i in obj)
    elif date_style == "mongo" and isinstance(obj, datetime.date):
        return datetime.datetime.combine(obj, datetime.datetime.min.time())
    elif date_style == "json" and isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()
    else:
        return obj

@dataclass
class Row():
    def to_dict(self):
        return asdict(self, dict_factory=lambda i: dict([(k, v) for k, v in i if is_valid(v)]))

    def to_mongo(self):
        return to_dict(self.to_dict, date_style="mongo")

    def to_pandas(self):
        """Convert object to Pandas Series"""
        import pandas as pd

        return pd.Series(self.to_dict())

    def to_json(self, *args, **kwargs):
        return json.dumps(to_dict(self.to_dict(), date_style="json"), *args, **kwargs)

    def fields(self):
        return fields(self)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"'{self.__class__.__name__}' object has not attribute '{key}'")

    def get(self, key, default):
        try:
            return self[key] or default
        except KeyError:
            return default

    def items(self):
        for f in fields(self):
            yield (f.name, getattr(self, f.name))

    def values(self):
        for f in fields(self):
            yield getattr(self, f.name)
    
    def keys(self):
        for f in fields(self):
            yield f.name

    def __bool__(self):
        return not all(v is None for v in self.__dict__.values())



from .autodetect import autodetect_format
from .iterparse import xml_iterparse


def sample_xml(
    in_file_obj: "io.RawIOBase", out_file_obj: "io.RawIOBase", n_records: int = 100
):
    chunk = in_file_obj.read(5000)
    format = autodetect_format(chunk)

    in_file_obj.seek(0)
    out_file_obj.write(format.preamble)
    for i, r in enumerate(xml_iterparse(in_file_obj, tag=format.item_tag)):
        out_file_obj.write(r)
        if i >= n_records:
            break
    out_file_obj.write(format.postamble)



import io
import re


def file_iterparse(
    in_file: io.RawIOBase, start: bytes, end: bytes = None
) -> "Iterable[bytes]":
    """
    Given an input file, and a start regex, yields individual records from the file
    Takes an optional end regex to eliminate unwanted or unnessary interstitial matter
    """
    buffer = bytearray()
    for event, chunk in file_event_parser(in_file, start, end):
        if event in ("start", "middle", "end"):
            buffer += chunk
        if event == "end":
            yield bytes(buffer)
            buffer = bytearray()


class NullObject(object):
    """A nothing object that returns null to all possible
    method calls"""

    def __getattr__(self, name: str) -> "Any":
        return self.null_function

    def null_function(self, *args, **kwargs):
        return None


def safe_search(string, regex):
    return regex.search(string) or NullObject()


def file_event_parser(
    in_file: io.RawIOBase, start: bytes, end: bytes = None, chunk_size=4000
) -> "Iterable[Tuple[str, bytes]]":
    """
    Given an input file, and a start and an optional end regex, yields
    "events" with chunks of the file.

    Inspired by etree.iterparse, yields back "events":
        None - preamble that isn't part of a record
        start - the start of a new record
        middle - continuation of a record
        end - the end of a record
    """
    if isinstance(start, (bytes, str)):
        if isinstance(start, str):
            start = start.encode()
        start = re.compile(start)
    if isinstance(end, (bytes, str)):
        if isinstance(end, str):
            end = end.encode()
        end = re.compile(end)

    chunk = bytearray()
    min_window_size = max((len(start.pattern), len(end.pattern) if end else 0))
    last_event = None
    while True:
        chunk_length = len(chunk)
        start_index = safe_search(chunk, start).start(0)
        end_index = safe_search(chunk, end).end(0) if end else None
        # Middle of Record
        if start_index is None and end_index is None:
            if last_event:
                yield ("middle", chunk)
                last_event = "middle"
            else:
                yield (None, chunk)
                last_event = None
            chunk = bytearray()

        # Start of Record
        elif (
            start_index is not None
            and end_index is None
            or (start_index and end_index and start_index < end_index)
        ):
            if last_event:
                yield ("end", chunk[:start_index])
                last_event = "end"
            else:
                yield (None, chunk[:start_index])
            yield ("start", chunk[start_index : start_index + len(start.pattern)])
            last_event = "start"
            chunk = chunk[start_index + len(start.pattern) :]
            recording = True

        # End of record
        elif (
            end_index is not None
            and start_index is None
            or (start_index and end_index and end_index < start_index)
        ):
            # It is implied that you are recording
            yield ("end", chunk[:end_index])
            last_event = "end"
            chunk = chunk[end_index:]
            last_event = None

        if chunk_length <= min_window_size:
            new_chunk = in_file.read(chunk_size)
            if not new_chunk:
                break
            chunk += new_chunk

    if last_event in ("middle", "start"):
        yield ("end", b"")



import lxml.html as ET
from lxml.etree import _Element
from yankee.base import schema
from yankee.base.deserializer import Deserializer

from .mixin import HtmlMixin

class Deserializer(HtmlMixin, Deserializer):
    pass

class Schema(HtmlMixin, schema.Schema):
    def load(self, obj):
        if isinstance(obj, _Element):
            return super().load(obj)
        elif isinstance(obj, str):
            return super().load(ET.fromstring(obj.encode()))
        elif isinstance(obj, bytes):
            return super().load(ET.fromstring(obj))


class PolymorphicSchema(HtmlMixin, schema.PolymorphicSchema):
    pass


class RegexSchema(HtmlMixin, schema.RegexSchema):
    pass

class ZipSchema(HtmlMixin, schema.ZipSchema):
    pass


import lxml.etree as ET
from yankee.util import clean_whitespace
from .accessor import html_accessor

class HtmlMixin(object):
    class Meta:
        accessor_function = html_accessor
        infer_keys = False

    def to_string(self, elem):
        if isinstance(elem, str):
            return elem
        elif isinstance(elem, ET._Comment):
            return clean_whitespace(elem.text, preserve_newlines=True)
        elif isinstance(elem, ET._Element):
            return clean_whitespace("".join(elem.itertext()), preserve_newlines=True)

    def convert_groupdict(self, dictionary):
        root = ET.Element("root")
        for k, v in dictionary.items():
            subel = ET.SubElement(root, k)
            subel.text = v
        return root

    @property
    def raw_text(self):
        return ET.tostring(self.raw, pretty_print=True).decode()



import io

from .iterparse import xml_iterparse

example_doc = """
<?xml version=1.0>
<collection>
<item>Some Text</item>
<item/>
<item/>
</collection>
<collection>
<item/>
</collection>"""


def test_xml_iterparse():
    counter = 0
    for el in xml_iterparse(io.BytesIO(example_doc.encode()), tag="item"):
        print(el)
        counter += 1
    assert counter == 4



import json
from copy import deepcopy
from itertools import chain

from .util import JsonEncoder
from .row import Row
from .util import resolve
from .util import to_dict
from .attrdict import AttrDict

class Collection:
    def __init__(self, iterable):
        self.iterable = iterable

    def __iter__(self):
        return iter(self.iterable)

    def __repr__(self):
        if hasattr(self, "iterable"):
            return f"Collection({repr(self.iterable)})"
        return super().__repr__()

    def __add__(self, other):
        return Collection(chain(self, other))

    def to_list(self):
        """Return a list of item objects from the Collection"""
        return ListCollection(self)

    def to_records(self, item_class=dict, collection_class=list):
        """Return a list of dictionaries containing item data in ordinary Python types
        Useful for ingesting into NoSQL databases
        """
        return to_dict(self, item_class, collection_class)

    def to_mongo(self):
        """Return a list of dictionaries containing MongoDB compatible datatypes
        """
        return to_dict(self, dict, list, convert_dates=True)

    def to_json(self, *args, **kwargs) -> str:
        """Convert objects to JSON format"""
        return json.dumps(list(self.to_records()), *args, cls=JsonEncoder, **kwargs)

    def to_pandas(self, annotate=list()):
        """Convert Collection into a Pandas DataFrame"""
        import pandas as pd

        list_of_series = list()
        for i in iter(self):
            try:
                series = i.to_pandas()
            except AttributeError:
                series = pd.Series(i)
            for a in annotate:
                series[a] = resolve(i, a)
            list_of_series.append(series)
        return pd.DataFrame(list_of_series)

    def explode(self, attribute, unpack=False, connector=".", prefix=True):
        """Implement an "explode" function for nested listed objects."""
        if unpack:
            return UnpackedCollection(ExplodedCollection(self, attribute), attribute, connector, prefix)
        else:
            return ExplodedCollection(self, attribute)

    def unpack(self, attribute, connector=".", prefix=True):
        """Implement an "unpack" function for nested single objects"""
        return UnpackedCollection(self, attribute, connector, prefix)

    # Values
    def values(self, *fields, **kw_fields):
        """Return a Collection that will return a Row object for each item with a subset of attributes
        positional arguments will result in Row objects where the fields match the field names on the item,
        keyword arguments can be used to rename attributes. When passed as key=field, the resulting dictionary will have key: item[field]
        """
        return ValuesCollection(self, *fields, **kw_fields)

    def values_list(self, *fields, flat=False, **kw_fields):
        """Return a Collection that will return tuples for each item with a subset of attributes.
        If only a single field is passed, the keyword argument "flat" can be passed to return a simple list"""
        return ValuesListCollection(self, *fields, flat=flat, **kw_fields)

class ListCollection(list, Collection):
    def __getitem__(self, sl):
        result = list(self)[sl]
        if isinstance(sl, slice):
            return ListCollection(result)
        else:
            return result

class ExplodedCollection(Collection):
    def __init__(self, iterable, attribute):
        self.iterable = iterable
        self.attribute = attribute

    def __iter__(self):
        for row in self.iterable:
            explode_field = resolve(row, self.attribute)
            for item in explode_field:
                new_row = row.to_dict()
                new_row[self.attribute] = item
                yield new_row


class UnpackedCollection(Collection):
    def __init__(self, iterable, attribute, connector=".", prefix=True):
        self.iterable = iterable
        self.attribute = attribute
        self.connector = connector
        self.prefix = prefix

    def item_key(self, k):
        if not self.prefix:
            return k
        else:
            return f"{self.attribute}{self.connector}{k}"

    def __iter__(self):
        for row in self.iterable:
            unpack_field = {self.item_key(k): v for k, v in resolve(row, self.attribute).items()}
            new_row = Row({**row, **unpack_field})
            del new_row[self.attribute]
            yield new_row


class ValuesCollection(Collection):
    def __init__(self, Collection, *arg_fields, fields=dict(), **kw_fields):
        self.Collection = Collection
        self.fields = {**{k: k for k in arg_fields}, **kw_fields, **fields}

    def __iter__(self):
        for item in self.Collection:
            yield AttrDict((k, resolve(item, v)) for k, v in self.fields.items())

    def __getitem__(self, sl):
        mger = deepcopy(self)
        new_mgr = mger.Collection.__getitem__(sl)
        return new_mgr


class ValuesListCollection(ValuesCollection):
    def __init__(self, Collections, *fields, flat=False, **kw_fields):
        super(ValuesListCollection, self).__init__(Collections, *fields, **kw_fields)
        self.flat = flat

    def __iter__(self):
        if self.flat and len(self.fields) > 1:
            raise ValueError("Flat only works with 1 field!")
        for row in super(ValuesListCollection, self).__iter__():
            data = tuple(row.values())
            yield data[0] if self.flat else data



from yankee.util import is_valid, inflect, update_class
from .accessor import python_accessor

class Deserializer(object):
    class Meta:
        accessor_function = python_accessor
        infer_keys = True
        output_style = "python"

    def __init__(self, data_key=None, many=False, required=False):
        self.data_key = data_key
        self.required = required
        self.many = many
        self.name = None
        self._build_meta()
        self.bind()

    def _build_meta(self):
        for c in self.__class__.mro():
            if not hasattr(c, "Meta"):
                continue
            for k in filter(lambda k: not k.startswith("_"), c.Meta.__dict__.keys()):
                if not hasattr(self.Meta, k):
                    setattr(self.Meta, k, getattr(c.Meta, k))
                
    def bind(self, name=None, parent=None):
        self.name = name
        self.parent = parent
        # Update Meta object
        if self.parent is not None:
            self.Meta = self.parent.Meta
        # Regenerate Accessor
        self.make_accessor()
        # Set Output Name
        if self.name is not None:
            self.output_name = inflect(self.name, style=self.Meta.output_style)
        return self

    def make_accessor(self):
        self.accessor = self.Meta.accessor_function(self.data_key, self.name, self.many, self.Meta)

    def load(self, obj):
        self.raw = obj
        pre_obj = self.pre_load(obj)
        loaded_obj = self.deserialize(pre_obj)
        loaded_obj = self.load_model(loaded_obj)
        return self.post_load(loaded_obj)

    def pre_load(self, obj):
        return obj

    def load_model(self, obj):
        return obj
    
    def deserialize(self, obj):
        return self.accessor(obj)

    def post_load(self, obj):
        return obj



import datetime

import pytest

from yankee import Schema, fields as f

doc1 = {
    "string": "Some String Data",
    "date_time": "2021-05-04T12:05",
    "date": "2021-05-04",
    "booleans": ["True", "true", "False", "false"],
    "float": 1.234,
    "int": 23,
    "exists": "Something",
    "name": {"part1": "George", "part2": "Burdell"},
    "random": "Some data",
    "address_line_1": "1234 Anywhere",
    "address_line_2": "Austin, TX 71234",
    "bad_string": "",
}

def to_obj(obj):
    if isinstance(obj, dict):
        return type("Dict", (object,), {k: to_obj(v) for k, v in obj.items()})
    elif isinstance(obj, list):
        return [to_obj(i) for i in obj]
    else:
        return obj

doc2 = to_obj(doc1)

class AddressField(f.Combine):
    address_line_1 = f.Str()
    address_line_2 = f.Str()

    def combine_func(self, obj):
        return f"{obj.address_line_1}\n{obj.address_line_2}"

class NameSchema(f.Combine):
    part1 = f.Str()
    part2 = f.Str()

    def combine_func(self, obj):
        return f"{obj.part1} {obj.part2}"

class SubSchema(Schema):
    string = f.Str()
    float = f.Float()

class ExampleSchema(Schema):
    string = f.Str()
    field = f.Field("string")
    date_time = f.DT()
    date = f.Date()
    booleans = f.List(f.Bool, data_key="booleans")
    float = f.Float()
    int = f.Int()
    exists = f.Exists()
    does_not_exist = f.Exists()
    name = NameSchema()
    sub = SubSchema(False)
    address = AddressField(False)
    bad_string = f.Str()



def test_fields_on_dict():
    schema = ExampleSchema()
    data = schema.load(doc1)
    assert data.string == "Some String Data"
    assert data.field == "Some String Data"
    assert data.date_time == datetime.datetime(2021, 5, 4, 12, 5)
    assert data.date == datetime.date(2021, 5, 4)
    assert data.booleans == [True, True, False, False]
    assert data.float - 1.234 < 0.001
    assert data.int == 23
    assert data.exists == True
    assert data.does_not_exist == False
    assert data.name == "George Burdell"
    assert data.sub.string == "Some String Data"
    assert data.address == "1234 Anywhere\nAustin, TX 71234"
    assert data.bad_string == None

def test_fields_on_obj():
    schema = ExampleSchema()
    data = schema.load(doc2)
    assert data.string == "Some String Data"
    assert data.field == "Some String Data"
    assert data.date_time == datetime.datetime(2021, 5, 4, 12, 5)
    assert data.date == datetime.date(2021, 5, 4)
    assert data.booleans == [True, True, False, False]
    assert data.float - 1.234 < 0.001
    assert data.int == 23
    assert data.exists == True
    assert data.does_not_exist == False
    assert data.name == "George Burdell"
    assert data.sub.string == "Some String Data"
    assert data.address == "1234 Anywhere\nAustin, TX 71234"
    assert data.bad_string == None

class JsonExampleSchema(Schema):
    class Meta:
        output_style = "json"
    string = f.Str()
    date_time = f.DT()
    date = f.Date()

def test_json_output_type():
    schema = JsonExampleSchema()
    data = schema.load(doc2)
    assert data.string == "Some String Data"
    assert data.dateTime == "2021-05-04T12:05:00"
    assert data.date == "2021-05-04"

class SecondSchema(Schema):
    first_schema = f.Nested("FirstSchema", data_key=False)

class FirstSchema(Schema):
    string = f.Str()

delayed_data_doc = {
    "string": "Some String"
}

def test_delayed_imports():
    schema = SecondSchema()
    data = schema.load(delayed_data_doc)
    assert data.to_dict() == {"first_schema": {"string": "Some String"}}

class ListSchema(Schema):
    l = f.List("ItemSchema")

class ItemSchema(Schema):
    i = f.Str()

delayed_list_doc = {
    "l": [
        {"i": "string1"},
        {"i":"string2"}
    ]
}

def test_list_field_by_string():
    schema = ListSchema()
    data = schema.load(delayed_list_doc)
    assert data.to_dict() == delayed_list_doc

def test_loads_model():
    from .fixtures.schema import ObjectSchema
    from .fixtures.model import Object
    schema = ObjectSchema()
    assert schema.__model__ == Object

def test_loads_nested_model():
    from .fixtures.nested.schema.example import ObjectSchema
    from .fixtures.nested.model.example import Object
    schema = ObjectSchema()
    assert schema.__model__ == Object


import datetime

import pytest
import ujson as json

from yankee.json.schema import Schema, fields as f, JsonPath

doc = {
    "string": "Some String Data",
    "dateTime": "2021-05-04T12:05",
    "date": "2021-05-04",
    "booleans": ["True", "true", "False", "false"],
    "float": 1.234,
    "int": 23,
    "exists": "Something",
    "name": {"part1": "George", "part2": "Burdell"},
    "random": "Some data",
    "addressLine1": "1234 Anywhere",
    "addressLine2": "Austin, TX 71234",
}

class AddressField(f.Combine):
    addressLine1 = f.Str()
    addressLine2 = f.Str()

    def combine_func(self, obj):
        return f"{obj.addressLine1}\n{obj.addressLine2}"

class NameSchema(f.Combine):
    part1 = f.Str()
    part2 = f.Str()

    def combine_func(self, obj):
        return f"{obj.part1} {obj.part2}"

class SubSchema(Schema):
    string = f.Str()
    float = f.Float()

class ExampleSchema(Schema):
    string = f.Str()
    string_path = f.Str(JsonPath("string"))
    date_time = f.DT()
    date = f.Date()
    booleans = f.List(f.Bool, data_key="booleans")
    float = f.Float()
    int = f.Int()
    exists = f.Exists()
    does_not_exist = f.Exists()
    name = NameSchema()
    sub = SubSchema(False)
    address = AddressField(False)
    


def test_fields():
    schema = ExampleSchema()
    doc_str = json.dumps(doc)
    data = schema.load(doc_str)
    assert data.string == "Some String Data"
    assert data.date_time == datetime.datetime(2021, 5, 4, 12, 5)
    assert data.date == datetime.date(2021, 5, 4)
    assert data.booleans == [True, True, False, False]
    assert data.float - 1.234 < 0.001
    assert data.int == 23
    assert data.exists == True
    assert data.does_not_exist == False
    assert data.name == "George Burdell"
    assert data.sub.string == "Some String Data"
    assert data.address == "1234 Anywhere\nAustin, TX 71234"

def test_json_path():
    doc = {'foo': [{'baz': 1}, {'baz': 2}]}
    class PathSchema(Schema):
        values = f.Int(JsonPath("foo[*].baz"))
    data = PathSchema().load(doc)
    assert data.values == 1

    class PathListSchema(Schema):
        values = f.List(f.Int, JsonPath("foo[*].baz"))
    data = PathListSchema().load(doc)
    assert data.values == [1, 2]


from .autodetect import autodetect_format

example_doc = """<?xml version="1.0" encoding="UTF-8" ?>
<uspat:PatentBulkData xsi:schemaLocation="urn:us:gov:doc:uspto:patent ../../main/resources/Schema/USPatent/Document/PatentBulkData_V8_0.xsd" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" com:st96Version="V3_1" com:ipoVersion="US_V8_0">
<uspat:PatentData com:st96Version="V4_0" com:ipoVersion="US_V9_0" xsi:schemaLocation="null null" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:tbl="http://www.oasis-open.org/tables/exchange/1.0" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:m="http://www.w3.org/1998/Math/MathML3" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <uspat:PatentCaseMetadata>
        <uscom:ApplicationNumberText uscom:electronicText="02022946">02022946</uscom:ApplicationNumberText>
        <uscom:ApplicationTypeCategory>Utility</uscom:ApplicationTypeCategory>
        <uspat:PartyBag/>
        <uspat:ApplicationConfirmationNumber>1141</uspat:ApplicationConfirmationNumber>
        <uscom:BusinessEntityStatusCategory>UNDISCOUNTED</uscom:BusinessEntityStatusCategory>
        <pat:InventionTitle>N/A</pat:InventionTitle>
        <uscom:OfficialFileLocationDate>2000-01-21</uscom:OfficialFileLocationDate>
    </uspat:PatentCaseMetadata>
</uspat:PatentData>
<uspat:PatentData com:st96Version="V4_0" com:ipoVersion="US_V9_0" xsi:schemaLocation="null null" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:tbl="http://www.oasis-open.org/tables/exchange/1.0" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:m="http://www.w3.org/1998/Math/MathML3" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <uspat:PatentCaseMetadata>
        <uscom:ApplicationNumberText uscom:electronicText="02000161">02000161</uscom:ApplicationNumberText>
        <uscom:ApplicationTypeCategory>Utility</uscom:ApplicationTypeCategory>
        <uspat:PartyBag/>
        <uspat:ApplicationConfirmationNumber>1014</uspat:ApplicationConfirmationNumber>
        <uscom:BusinessEntityStatusCategory>UNDISCOUNTED</uscom:BusinessEntityStatusCategory>
        <pat:InventionTitle>N/A</pat:InventionTitle>
        <uscom:OfficialFileLocationDate>2000-01-21</uscom:OfficialFileLocationDate>
    </uspat:PatentCaseMetadata>
</uspat:PatentData>
</uspat:PatentBulkData"""


def test_autodetect():
    format = autodetect_format(example_doc.encode())
    assert format.item_tag == "uspat:PatentData"
    assert format.list_tag == "uspat:PatentBulkData"
    assert (
        format.preamble
        == '<?xml version="1.0" encoding="UTF-8" ?>\n<uspat:PatentBulkData xsi:schemaLocation="urn:us:gov:doc:uspto:patent ../../main/resources/Schema/USPatent/Document/PatentBulkData_V8_0.xsd" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" com:st96Version="V3_1" com:ipoVersion="US_V8_0">'
    )
    assert format.postamble == "</uspat:PatentBulkData>"
    assert format.multidoc == False
    assert format.namespaces == {
        "pat": "http://www.wipo.int/standards/XMLSchema/ST96/Patent",
        "uscom": "urn:us:gov:doc:uspto:common",
        "uspat": "urn:us:gov:doc:uspto:patent",
        "com": "http://www.wipo.int/standards/XMLSchema/ST96/Common",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    }


example_multidoc = """<?xml version="1.0" encoding="UTF-8" ?>
<uspat:PatentBulkData xsi:schemaLocation="urn:us:gov:doc:uspto:patent ../../main/resources/Schema/USPatent/Document/PatentBulkData_V8_0.xsd" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" com:st96Version="V3_1" com:ipoVersion="US_V8_0">
<uspat:PatentData com:st96Version="V4_0" com:ipoVersion="US_V9_0" xsi:schemaLocation="null null" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:tbl="http://www.oasis-open.org/tables/exchange/1.0" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:m="http://www.w3.org/1998/Math/MathML3" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <uspat:PatentCaseMetadata>
        <uscom:ApplicationNumberText uscom:electronicText="02022946">02022946</uscom:ApplicationNumberText>
        <uscom:ApplicationTypeCategory>Utility</uscom:ApplicationTypeCategory>
        <uspat:PartyBag/>
        <uspat:ApplicationConfirmationNumber>1141</uspat:ApplicationConfirmationNumber>
        <uscom:BusinessEntityStatusCategory>UNDISCOUNTED</uscom:BusinessEntityStatusCategory>
        <pat:InventionTitle>N/A</pat:InventionTitle>
        <uscom:OfficialFileLocationDate>2000-01-21</uscom:OfficialFileLocationDate>
    </uspat:PatentCaseMetadata>
</uspat:PatentData>
</uspat:PatentBulkData
<?xml version="1.0" encoding="UTF-8" ?>
<uspat:PatentBulkData xsi:schemaLocation="urn:us:gov:doc:uspto:patent ../../main/resources/Schema/USPatent/Document/PatentBulkData_V8_0.xsd" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" com:st96Version="V3_1" com:ipoVersion="US_V8_0">
<uspat:PatentData com:st96Version="V4_0" com:ipoVersion="US_V9_0" xsi:schemaLocation="null null" xmlns:uspat="urn:us:gov:doc:uspto:patent" xmlns:tbl="http://www.oasis-open.org/tables/exchange/1.0" xmlns:pat="http://www.wipo.int/standards/XMLSchema/ST96/Patent" xmlns:uscom="urn:us:gov:doc:uspto:common" xmlns:m="http://www.w3.org/1998/Math/MathML3" xmlns:com="http://www.wipo.int/standards/XMLSchema/ST96/Common" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <uspat:PatentCaseMetadata>
        <uscom:ApplicationNumberText uscom:electronicText="02022946">02022946</uscom:ApplicationNumberText>
        <uscom:ApplicationTypeCategory>Utility</uscom:ApplicationTypeCategory>
        <uspat:PartyBag/>
        <uspat:ApplicationConfirmationNumber>1141</uspat:ApplicationConfirmationNumber>
        <uscom:BusinessEntityStatusCategory>UNDISCOUNTED</uscom:BusinessEntityStatusCategory>
        <pat:InventionTitle>N/A</pat:InventionTitle>
        <uscom:OfficialFileLocationDate>2000-01-21</uscom:OfficialFileLocationDate>
    </uspat:PatentCaseMetadata>
</uspat:PatentData>
</uspat:PatentBulkData"""


def test_autodetect_on_multidoc():
    format = autodetect_format(example_multidoc.encode())
    assert format.multidoc == True



import pytest
import datetime
from dateutil.tz import tzutc

from .fields import * #: noqa

class TestDateTime():
    def test_dt_string(self):
        result = DateTime(dt_format="%m/%d/%Y").deserialize("03/15/2019")
        assert result == datetime.datetime(2019, 3, 15)

    def test_isoformat(self):
        dt = datetime.datetime.now()
        result = DateTime().load(dt.isoformat())
        assert result == dt

    def test_bad_isoformat(self):
        field = DateTime()
        result = field.load("2017-04-18T04:00:00Z")
        assert result == datetime.datetime(2017, 4, 18, 4, 0, tzinfo=tzutc())

class TestDate():
    def test_dt_string(self):
        result = Date(dt_format="%m/%d/%Y").load("03/15/2019")
        assert result == datetime.date(2019, 3, 15)

    def test_isoformat(self):
        dt = datetime.datetime.now().date()
        result = Date().load(dt.isoformat())
        assert result == dt

    def test_bad_isoformat(self):
        field = Date()
        result = field.load("2017-04-18T04:00:00Z")
        assert result == datetime.date(2017, 4, 18)

class TestInt():
    def test_int(self):
        field = Int()
        assert field.load("15") == 15

    def test_missing_int(self):
        field = Int()
        assert field.load("") == None
        assert field.load(None) == None

    def test_bad_int(self):
        field = Int()
        with pytest.raises(ValueError):
            field.load("Some text")

class TestFloat():
    def test(self):
        field = Float()
        assert field.load("15.24") - 15.24 < 0.001

    def test_missing(self):
        field = Float()
        assert field.load("") == None
        assert field.load(None) == None

    def test_bad(self):
        field = Float()
        with pytest.raises(ValueError):
            field.load("Some text")

class TestList():
    def test(self):
        data = [1, 2, 3]
        field = List(item_schema=Str())
        result = field.load(data)
        assert result == ["1", "2", "3"]

    def test_missing(self):
        field = List(item_schema=Str())
        result = field.load(None)
        assert result == list()


from __future__ import annotations
import typing
import datetime
import re
import importlib
import warnings

from dateutil.parser import parse as parse_dt, isoparse

from yankee.util import AttrDict, clean_whitespace, is_valid
from yankee.data.collection import ListCollection

from .deserializer import Deserializer
from .schema import Schema
from functools import partial


class Field(Deserializer):
    output_type = typing.Any
    def __init__(self, *args, default=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = default
    
    def load(self, obj):
        result = super().load(obj)
        return result if is_valid(result) else self.default

class String(Field):
    output_type = str
    def __init__(self, *args, formatter=None, null_value=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.formatter = formatter or partial(clean_whitespace, preserve_newlines=True)
        self.null_value = null_value

    def deserialize(self, elem) -> "Optional[str]":
        elem = super().deserialize(elem)
        if elem is None or elem == "" or elem == self.null_value:
            return None
        else:
            return self.formatter(self.to_string(elem)) if self.formatter else self.to_string(elem)

    def to_string(self, elem): # Abstracted out since XML requires a function call
        return str(elem)


class DateTime(String):
    output_type = datetime.datetime
    def __init__(self, *args, dt_format=None, **kwargs):
        super().__init__(*args, **kwargs)
        if dt_format:
            self.parse_date = lambda s: datetime.datetime.strptime(s, dt_format)

    def parse_date(self, text:str):
        try:
            return isoparse(text)
        except ValueError:
            return parse_dt(text)

    def deserialize(self, elem) -> "Optional[datetime.datetime]":
        string = super(DateTime, self).deserialize(elem)
        return self.parse_date(string) if string else None

    def post_load(self, obj):
        if self.Meta.output_style == "json":
            return obj.isoformat()
        return obj

class Date(DateTime):
    output_type = datetime.date
    def deserialize(self, elem) -> "Optional[datetime.date]":
        date_time = super().deserialize(elem)
        return date_time.date() if date_time else None


class Boolean(String):
    output_type = bool
    def __init__(
        self, *args, true_value="true", case_sensitive=False, allow_none=True, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.true_value = true_value
        self.case_sensitive = case_sensitive
        self.allow_none = allow_none
        if not self.case_sensitive:
            self.true_value = self.true_value.lower()

    def deserialize(self, elem) -> "Optional[bool]":
        string = super(Boolean, self).deserialize(elem)
        if string is None or string == '':
            return None if self.allow_none else False
        if not self.case_sensitive:
            string = string.lower()
        return string == self.true_value


class Float(String):
    output_type = float
    def deserialize(self, elem) -> "Optional[float]":
        string = super(Float, self).deserialize(elem)
        return float(string) if string is not None else None


class Integer(String):
    output_type = int
    def deserialize(self, elem) -> "Optional[int]":
        string = super(Integer, self).deserialize(elem)
        return int(string) if string is not None else None


class Exists(Field):
    output_type = bool
    def deserialize(self, elem) -> bool:
        obj = super(Exists, self).deserialize(elem)
        return obj is not None


class Const(Field):
    def __init__(self, const, output_type=None, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.const = const
        if output_type is None:
            warnings.warn(f"Constant has unassigned output type - cannot infer output schema without constant type information")
        self.output_type = output_type

    def deserialize(self, elem) -> "Any":
        return self.const
    

# Multiple Value Fields
    
class Nested(Schema):
    output_type = dict
    def __init__(self, schema, *args, **kwargs):
        self._schema = schema
        self._args = args
        self._kwargs = kwargs

    # Deserialize Methods
    def bind(self, name=None, parent=None):
        super().bind(name=name, parent=parent)
        if isinstance(self._schema, str):
            *module, _schema = self._schema.split(".")
            module = ".".join(module) or parent.__module__
            schema_class = getattr(importlib.import_module(module), _schema)
            self._schema = schema_class(*self._args, **self._kwargs)
        self._schema.bind(name, parent)

    def make_accessor(self, *args, **kwargs):
        if isinstance(self._schema, str):
            return
        return self._schema.make_accessor(*args, **kwargs)

    def load(self, obj):
        return self._schema.load(obj)

class List(Field):
    output_type = list
    def __init__(self, item_schema, data_key=None, **kwargs):
        kwargs['many'] = True
        self.item_schema = item_schema
        if callable(self.item_schema):
            self.item_schema = item_schema()
        super().__init__(data_key, **kwargs)

    def bind(self, name=None, schema=None):
        super().bind(name, schema)
        if isinstance(self.item_schema, str):
            *module, _schema = self.item_schema.split(".")
            if module:
                module = ".".join(module)
            elif schema is not None:
                module = schema.__module__
            else:
                return
            self.item_schema = getattr(importlib.import_module(module), self.item_schema)()
        self.item_schema.bind(None, schema)

    def load(self, obj):
        obj = self.deserialize(obj)
        if not obj:
            return ListCollection()
        obj_gen = (self.item_schema.load(i) for i in obj)
        return ListCollection(o for o in obj_gen if is_valid(o))

class Dictionary(List):
    output_type = dict
    """Converts a list of items into a dictionary based on
    the key and value fields passed to it.
    """
    
    def __init__(self, data_key, key:Field, value:Field, **kwargs):
        self.key = key
        self.value = value
        return super().__init__(Field(), data_key, **kwargs)
    
    def load(self, obj):
        obj = self.deserialize(obj)
        if not obj:
            return AttrDict()
        return AttrDict((self.key.load(i), self.value.load(i)) for i in obj)

# String Parsing Fields

class DelimitedString(String):
    output_type = list
    def __init__(self, item_schema, data_key=None, delimeter=",", **kwargs):
        self.item_schema = item_schema
        if not isinstance(delimeter, re.Pattern):
            delimeter = re.compile(delimeter)
        self.delimeter = delimeter
        if callable(self.item_schema):
            self.item_schema = item_schema()
        super().__init__(data_key=data_key, **kwargs)

    def bind(self, name=None, schema=None):
        super().bind(name, schema)
        self.item_schema.bind(None, schema)

    def deserialize(self, obj):
        obj = super().deserialize(obj)
        if obj is None:
            return None
        objs = (self.item_schema.load(o) for o in self.delimeter.split(obj))
        return [o for o in objs if is_valid(o)]

# Schema-Like Fields

class Combine(Schema):
    """Can have fields like a schema that are then
    passed as an object to a combine function that
    transforms it to a single string value"""
    output_type = str

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        for field in self.fields.values():
            field.output_name = field.name
        self.__model__ = self.make_dataclass()

    def load(self, obj):
        loaded_obj = super().load(obj)
        return self.combine_func(loaded_obj)

    def combine_func(self, obj):
        raise NotImplementedError("Must be implemented in subclass")


class Alternative(Schema):
    """There may be a piece of data that has different names
    in different contexts. This has fields like a schema, then
    passes as a value the first non-empty or non-null result"""

    def deserialize(self, et_elem):
        obj = super().deserialize(et_elem)
        return next((v for v in obj.values() if is_valid(v)), None)


# Aliases
Str = String
DT = DateTime
Bool = Boolean
Int = Integer
Alt = Alternative
Dict = Dictionary


# XPath Utilites

# XPath Node Set Operations
def xpath_intersection(ns1, ns2):
    """The Kaysian Method for XPath 1.0 intersections
    Do not include leading . in ns1 and ns2"""
    return f"{ns1}[count(.|{ns2})=count({ns2})]"

def xpath_difference(ns1, ns2):
    """The Kaysian Method for XPath 1.0 differences
    Do not include leading . in ns1 and ns2"""
    return f"{ns1}[count({ns1}) != count({ns2})]"


import datetime

import lxml.etree as ET
import pytest

from yankee.xml.schema import Schema, RegexSchema, fields as f, XPath, CSS

from .fields import *

test_doc = """
<testdoc>
    <string>Some String Data</string>
    <regex>data_a;data_b</regex>
    <date_time>2021-05-04T12:05</date_time>
    <date>2021-05-04</date>
    <booleans>
        <bool>True</bool>
        <bool>true</bool>
        <bool>False</bool>
        <bool>false</bool>
        <bool/>
    </booleans>
    <!-- A Comment -->
    <float>1.234</float>
    <int>23</int>
    <exists>Something</exists>
    <part1>George</part1>
    <part2>Burdell</part2>
    <random>Some data</random>
    <csv>name1,name2,name3</csv>
    <dict>
        <item name="key1">value1</item>
        <item name="key2">value2</item>
    </dict>
    <zip>
        <first_name>
            <name>Parker</name>
            <name>Peter</name>
        </first_name>
        <last_name>
            <name>Hancock</name>
            <name>Parker</name>
        </last_name>
    </zip>
</testdoc>
""".strip()


tree = ET.fromstring(test_doc.encode())

class NameCombineSchema(f.Combine):
    first_name = f.Str(".//first_name/name")
    last_name = f.Str(".//last_name/name")

    def combine_func(self, obj):
        return f"{obj.first_name} {obj.last_name}"

class NameSchema(Combine):
    part1 = f.Str(".//part1")
    part2 = f.Str(".//part2")

    def combine_func(self, obj):
        return f"{obj.part1} {obj.part2}"

class RegexExample(RegexSchema):
    a = f.Str("./a")
    b = f.Str("./b")
    
    __regex__ = r"(?P<a>[^;]+);(?P<b>[^;]+)"


class ExampleSchema(Schema):
    comment = f.Str("./comment()")
    string = f.Str("./string")
    other_string = f.Str("./string/text()")
    date_time = f.DT("./date_time")
    date = f.Date("./date")
    booleans = f.List(Bool, "./booleans/bool")
    float = f.Float("./float")
    int = f.Int("./int")
    exists = f.Exists("./exists")
    does_not_exist = f.Exists("./does_not_exist")
    name = NameSchema()
    regex = RegexExample("./regex")
    bad_regex = RegexExample("./missing")
    gone = f.Str("./nonexistent_path")
    csv = f.DelimitedString(f.Str(), data_key="./csv", delimeter=",")
    dict = f.Dict("./dict/item", key=f.Str(".//@name"), value=f.Str())


def test_fields():
    d = ExampleSchema()
    data = d.load(test_doc).to_dict()
    assert data["string"] == "Some String Data"
    assert data["date_time"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]
    assert data["float"] - 1.234 < 0.001
    assert data["int"] == 23
    assert data["exists"] == True
    assert data["does_not_exist"] == False
    assert data["name"] == "George Burdell"
    assert data['regex']['a'] == 'data_a'
    assert data['regex']['b'] == 'data_b'
    assert "gone" not in data
    assert data['csv'] == ['name1', 'name2', 'name3']
    assert data['dict'] == {
        "key1": "value1",
        "key2": "value2",
    }
    assert data['comment'] == "A Comment"

ns_test_doc = """
<?xml version='1.0' encoding='UTF-8'?>
<h:testDoc xmlns:h="http:www.w3.org/TR/html4/">
    <h:string>Some String Data</h:string>
    <h:date_time>2021-05-04T12:05</h:date_time>
    <h:date>2021-05-04</h:date>
    <h:booleans>
        <bool>True</bool>
        <bool>true</bool>
        <bool>False</bool>
        <bool>false</bool>
    </h:booleans>
</h:testDoc>
""".strip()

ns_tree = ET.fromstring(ns_test_doc.encode())

class NsSchema(Schema):
    class Meta():
        namespaces = {
            "h": "http:www.w3.org/TR/html4/"
        }
    string = f.Str("./h:string")
    date_time = f.DT("./h:date_time")
    date = f.Date("./h:date")
    booleans = f.List(Bool, "./h:booleans/bool")
   
def test_ns_fields():
    d = NsSchema()
    data = d.load(ns_tree).to_dict()
    assert data["string"] == "Some String Data"
    assert data["date_time"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]

def test_missing_field():
    doc = b"""<doc><name><first_name>Parker</first_name></name></doc>"""
    tree = ET.fromstring(doc)

    class NameSchema(Schema):
        first_name = f.Str("./first_name")
        last_name = f.Str("./last_name")

    class ExampleSchema(Schema):
        name = NameSchema("./name_1")

    result = ExampleSchema().load(tree).to_dict()
    assert result == dict()

def test_literal_xpath():
    doc = b"<html><a>data 1</a><a>data 2</a></html>"

    class XpathSchema(Schema):
        a = f.Str(XPath(".//a"))
    
    data = XpathSchema().load(doc)
    assert data.a == "data 1"

    class XpathListSchema(Schema):
        a = f.List(f.Str, XPath(".//a"))
    
    data = XpathListSchema().load(doc)
    assert data.a == ["data 1", "data 2"]

def test_literal_css():
    doc = b"<html><a>data 1</a><a>data 2</a></html>"

    class CssSchema(Schema):
        a = f.Str(CSS("a"))
    
    data = CssSchema().load(doc)
    assert data.a == "data 1"

    class CssListSchema(Schema):
        a = f.List(f.Str, CSS("a"))
    
    data = CssListSchema().load(doc)
    assert data.a == ["data 1", "data 2"]


from yankee.base import schema
import ujson as json
from .mixin import JsonMixin


class Schema(JsonMixin, schema.Schema):
    def load(self, obj):
        if isinstance(obj, dict):
            return super().load(obj)
        elif isinstance(obj, str):
            return super().load(json.loads(obj))
        elif isinstance(obj, bytes):
            return super().load(json.loads(obj.decode()))

class PolymorphicSchema(JsonMixin, schema.PolymorphicSchema):
    pass

class RegexSchema(JsonMixin, schema.RegexSchema):
    pass

class ZipSchema(JsonMixin, schema.ZipSchema):
    pass


import lxml.etree as ET
from yankee.base import fields
from yankee.util import clean_whitespace

from .mixin import XmlMixin


class Field(XmlMixin, fields.Field):
    pass


class String(XmlMixin, fields.String):
    pass


class DateTime(XmlMixin, fields.DateTime):
    pass


class Date(XmlMixin, fields.Date):
    pass


class Boolean(XmlMixin, fields.Boolean):
    pass


class Float(XmlMixin, fields.Float):
    pass


class Integer(XmlMixin, fields.Integer):
    pass


class Exists(XmlMixin, fields.Exists):
    pass


class Const(XmlMixin, fields.Const):
    pass


class List(XmlMixin, fields.List):
    pass

class Dict(XmlMixin, fields.Dict):
    pass

class Combine(XmlMixin, fields.Combine):
    pass

class Alternative(XmlMixin, fields.Alternative):
    pass

class DelimitedString(XmlMixin, fields.DelimitedString):
    pass

class Nested(XmlMixin, fields.Nested):
    pass

class TailField(fields.Field):
    """Field to retreive tail text"""

    def load(self, obj):
        return super().load(obj)

    def deserialize(self, obj):
        return clean_whitespace(super().deserialize(obj).tail)

# Aliases
Str = String
DT = DateTime
Bool = Boolean
Int = Integer
Alt = Alternative



import logging

import lxml.etree as ET
from file_utils import file_iterparse


class XmlProcessor(object):
    parser = None
    record_tag = None
    dtd_resolver = ET.Resolver

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        # Expand namespace declarations if present
        if self.record_tag and ":" in self.record_tag:
            ns, tag = self.record_tag.split(":")
            self.record_tag = f"{{{self.schema.Meta['ns'][ns]}}}{tag}"

    def process(self, file_obj, meta=None):
        parse_element = self._parse_element
        if self.multidoc:
            for el in self._process_multidoc(file_obj):
                yield parse_element(el, meta)
        else:
            for el in self._process_normal(file_obj):
                yield parse_element(el, meta)

    def _parse_element(self, el, meta):
        r = self.parser.deserialize(el)
        if meta is not None:
            r["meta"] = meta
        return r

    def _process_multidoc(self, file_obj):
        start = r"<\?xml".encode("utf-8")
        end = r"\n(?=\<\?xml)".encode("utf-8")
        xml_parser = ET.XMLParser(load_dtd=True, recover=True)
        xml_parser.resolvers.add(self.dtd_resolver())
        for r in file_iterparse(file_obj, start, end):
            yield xml_parser.fromstring(r)

    def _process_normal(self, file_obj):
        et_gen = ET.iterparse(
            file_obj, load_dtd=True, recover=True, tag=self.record_tag, events=("end",)
        )
        et_gen.resolvers.add(self.dtd_resolver())
        for _, el in et_gen:
            yield el
            el.clear()


def parse_multidoc(file_obj, xml_parser=None):
    start = r"<\?xml".encode("utf-8")
    end = r"\n(?=\<\?xml)".encode("utf-8")
    xml_parser = xml_parser or ET.XMLParser()
    for r in file_iterparse(file_obj, start, end):
        yield xml_parser.fromstring(r)


def parse_xml_file(file_obj, record_tag):
    for _, el in ET.iterparse(file_obj, tag=record_tag):
        yield el
        el.clear()



import re

class NullObject(object):
    """A nothing object that returns null to all possible
    method calls"""

    def __getattr__(self, name: str) -> "Any":
        return self.null_function

    def null_function(self, *args, **kwargs):
        return None


def safe_search(string, regex):
    return regex.search(string) or NullObject()

def safe_read(in_f, chunk_size):
    chunk = in_f.read(chunk_size)
    while True:
        nxt_chunk = in_f.read(chunk_size)
        if not nxt_chunk:
            yield chunk
            chunk = nxt_chunk
            continue
        end_tag = nxt_chunk.find(b">")
        yield chunk + nxt_chunk[:end_tag]
        chunk = nxt_chunk[end_tag:]
        
def iter_file(in_f, start_regex, end_regex, chunksize):
    pending = False
    chunk = bytearray()
    chunk_iter = safe_read(in_f, chunksize)
    chunk += next(chunk_iter)

    while True:
        if len(chunk) < chunksize:
            new_chunk = next(chunk_iter)
            chunk += new_chunk
        start = safe_search(chunk, start_regex).start(0)
        end = safe_search(chunk, end_regex).end(0)
        # Initialization - skip to first record
        if not pending and not start:
            chunk = bytearray()
        # Complete record within chunk
        elif start and end:
            yield ("start", chunk[start:end])
            yield ("end", bytearray())
            chunk = chunk[end:]
        # start detected, no open preceding item
        elif start and not pending:
            yield ("start", chunk[start:])
            chunk = bytearray()
        # end detected, open preceding item
        elif end and pending:
            yield ("end", chunk[:end])
            pending = False
            chunk = chunk[end:]
        # no start or end detected, but pending record
        elif not start and not end and pending:
            yield ("middle", chunk)
            chunk = bytearray()
        else:
            raise ValueError("Uncaught parser condition!")
        if not chunk and not new_chunk:
            break
            
def iter_record(in_f, start, end, chunksize):
    record = bytearray()
    for event, chunk in iter_file(in_f, start, end, chunksize):
        if event == "start":
            record = chunk
        elif event == "middle":
            record += chunk
        elif event == "end":
            record += chunk
            yield bytes(record).strip()
            record = bytearray()

chunk_size = 30000

def xml_iterparse(file_obj: "io.RawBytesIO", tag=None, chunksize=chunk_size):
    start_re = re.compile(f"<{tag}".encode())
    end_re = re.compile(f"</{tag}>|{tag}/>".encode())
    yield from iter_record(file_obj, start_re, end_re, chunksize)


import datetime
from collections import abc
from yankee.util import camelize, underscore, AttrDict


def jsonify(obj):
    if isinstance(obj, abc.Mapping):
        return AttrDict((camelize(k), jsonify(v)) for k, v in obj.items())
    elif isinstance(obj, abc.Sequence) and not isinstance(obj, (str, bytes)):
        return [jsonify(i) for i in obj]
    elif isinstance(obj, datetime.datetime):
        return obj.isoformat()
    elif isinstance(obj, datetime.date):
        return datetime.datetime.combine(obj, datetime.datetime.min.time()).isoformat()
    else:
        return obj

def pythonify(obj):
    if isinstance(obj, abc.Mapping):
        return AttrDict((underscore(k), pythonify(v)) for k, v in obj.items())
    elif isinstance(obj, abc.Sequence) and not isinstance(obj, (str, bytes)):
        return [pythonify(i) for i in obj]
    else:
        return obj


from yankee.util import camelize
from collections.abc import Sequence, Mapping

def do_nothing(obj):
    return obj

def python_accessor(data_key, name, many, meta):
    if data_key is False or (data_key is None and name is None):
        return do_nothing
    data_key = data_key or name
    if not isinstance(data_key, str):
        return data_key
    key_segments = tuple(int(s) if s.isdigit() else s for s in data_key.split("."))
    def accessor_func(obj):
        result = obj
        for seg in key_segments:
            if isinstance(obj, Sequence):
                try:
                    result = result[seg]
                except IndexError:
                    result = None
            elif isinstance(result, Mapping):
                result = result.get(seg, None)
            else:
                result = getattr(result, seg, None)
        return result
    return accessor_func






import io

import lxml.etree as ET

from .aps import aps_iterator, aps_record_to_xml

sample_doc = """
PATN
WKU  D04520599
APN  1348817
TTL  Pair of medium-heeled open-back shoes
ISD  20011218
TBL  ==========A FORMATTED TABLE==========
     -------------------------------------
INVT
NAM  Jacobs; Marc
ASSG
NAM  Louis Vuitton Malletier, S.A.
CTY  Paris
PRIR
CNT  FRX
APD  20000307
APN  00 3859
CLAS
OCL  D 2926
XCL   D2925
EDF  7
ICL  0204
UREF
PNO  D39412
ISD  19080700
NAM  Ostrowsky
OCL  D 2919
UREF
PNO  D95540
ISD  19350500
NAM  Perugia
OCL  D 2926
UREF
PNO  D190164
ISD  19610400
NAM  Lacek
OCL  D 2932""".strip()


def test_aps_conversion_from_string():
    result = aps_record_to_xml(sample_doc)
    tree = ET.fromstring(result)
    assert tree.find("./APN").text == "1348817"
    assert tree.find("./CLAS/OCL").text == "D 2926"
    assert len(tree.findall("./UREF")) == 3


def test_aps_conversion_from_bytes():
    result = aps_record_to_xml(sample_doc.encode())
    tree = ET.fromstring(result)
    assert tree.find("./APN").text == "1348817"
    assert tree.find("./CLAS/OCL").text == "D 2926"
    assert len(tree.findall("./UREF")) == 3


sample_docs = """
PATN
WKU  D04520599
PATN
WKU  101231241
PATN
WKU  123125123""".strip()


def test_aps_iterator():
    counter = 0
    for record in aps_iterator(io.BytesIO(sample_docs.encode()), record_tag="PATN"):
        tree = ET.fromstring(record)
        assert tree.find("./WKU").text
        counter += 1
    assert counter == 3



import io

from ..iterparse import file_iterparse

test_doc = """
<record>
<record>
<record>
<record>
"""


def test_file_iterparse_with_record_start():
    records = list(file_iterparse(io.BytesIO(test_doc.encode()), "<"))
    assert len(records) == 4



import logging
import re
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class XmlFormat(object):
    item_tag: bytes
    list_tag: bytes
    preamble: bytes
    postamble: bytes
    namespaces: "Dict[str, str]"
    multidoc: bool


xml_doc_re = re.compile(r"\<\?xml".encode("utf-8"))
preamble_re = re.compile(
    r"(?P<preamble>^\<\?xml .*?>\s+<(?P<list_tag>[^\s]+).*?\>)\s*\<(?P<item_tag>[^\s]+).*?\>".encode(
        "utf-8"
    )
)


def get_namespaces(preamble):
    namespace_re = re.compile(r'xmlns:(?P<ns>[^=]+)="(?P<url>[^"]+)')
    namespace_matches = [m.groupdict() for m in namespace_re.finditer(preamble)]
    return {m["ns"]: m["url"] for m in namespace_matches}


def autodetect_format(chunk):
    """This function automatically detects the formatting of an XML
    record-style document, so long as it is fed a chunk that
    contains the doc declaration, wrapper element, and the opening tag
    of the first record
    """

    match = preamble_re.search(chunk).groupdict()
    preamble = match["preamble"].decode()
    list_tag = match["list_tag"].decode()
    item_tag = match["item_tag"].decode()
    postamble = f"</{list_tag}>"

    namespaces = get_namespaces(preamble)

    xml_docs = xml_doc_re.findall(chunk)
    multidoc = len(xml_docs) > 1
    return XmlFormat(item_tag, list_tag, preamble, postamble, namespaces, multidoc)



import re
import dataclasses as dc
import importlib
from yankee.util import is_valid, AttrDict, clean_whitespace, unzip_records

from yankee.data.row import Row
from .deserializer import Deserializer
from yankee.data.collection import ListCollection, Collection

class Schema(Deserializer):
    output_type = dict
    __model__ = None
    def __init__(
        self,
        *args,
        flatten=False,
        prefix=False,
        **kwargs
    ):
        self.flatten = flatten
        self.prefix = prefix
        super().__init__(*args, **kwargs)
        self.bind()

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        # Make sure that fields are grabbed from superclasses as well
        fields = self.get_fields()
        for name, field in fields.items():
            if self.prefix:
                field.bind(f"{self.name}_{name}", self)
            else:
                field.bind(name, self)
        self.fields = fields
        self.get_model()

    def get_fields(self):
        class_fields = list()
        for c in reversed(self.__class__.mro()):
            class_fields += [
                (k, v) for k, v in c.__dict__.items() if isinstance(v, Deserializer)
            ]
        return dict(class_fields)

    def get_model(self):
        # Model class is expressly there
        if not isinstance(self.__model__, str) and self.__model__ is not None:
            return
        # Model should be obtained from a string
        model_str = self.__model__ or self.__class__.__name__.replace("Schema", "")
        *module, _model = model_str.rsplit(".", 1)
        module = self.__class__.__module__.replace(".schema", ".model")
        try:
            self.__model__ = getattr(importlib.import_module(module), _model)
        except (ImportError, AttributeError):
            self.__model__ = self.make_dataclass()
        

    def deserialize(self, obj) -> "Dict":
        output = AttrDict()
        obj = self.accessor(obj)
        for key, field in self.fields.items():
            value = field.load(obj)
            # If there is no value, don't include anything in the output dictionary
            if not is_valid(value):
                continue
            # If the value isn't a dict, or there's not flatten directive, add and continue
            if not isinstance(value, dict) or not getattr(field, "flatten", False):
                output[field.output_name] = value
                continue
            # Merge in flattened fields
            output.update(value)
        return output

    def load_model(self, obj):
        try:
            return self.__model__(**obj)
        except TypeError:
            return None

    def make_field(self, t):
        if t == list:
            return dc.field(default_factory=ListCollection)
        else:
            return dc.field(default=None)

    def make_dataclass(self):
        fields = list(
            (f.output_name, f.output_type, self.make_field(f.output_type))
            for f in self.fields.values()
            )
        dataclass = dc.make_dataclass(
            cls_name=self.__class__.__name__.replace("Schema", ""),
            fields=fields,
            bases=(Row,)
        )
        return dataclass

    def load_batch(self, objs):
        return Collection(self.load(o) for o in objs)


class PolymorphicSchema(Schema):
    def bind(self, name=None):
        super().bind(self, name)
        for schema in self.schemas:
            schema.bind(name)

    def choose_schema(self, obj):
        raise NotImplementedError("Must be implemented in subclass!")

    def deserialize(self, raw_obj) -> "Dict":
        # Get the key one time only, rather than
        # on both deserializing the selector obj
        # and the final output
        obj = super().deserialize(obj)
        schema = self.choose_schema(obj)
        return schema.deserialize(raw_obj)

class RegexSchema(Schema):
    """
    This schema type allows for using a regex to pull data
    out of a string, and then treat it like a schema
    """
    __regex__ = None
    
    def __init__(self, *args, **kwargs):
        self._regex = re.compile(self.__regex__)
        super().__init__(*args, **kwargs)
        
    def deserialize(self, obj):
        obj = self.accessor(obj)
        if obj is None:
            return dict()
        text = clean_whitespace(self.to_string(obj))
        match = self._regex.search(text)
        if match is None:
            return dict()
        obj = self.convert_groupdict(match.groupdict())
        output = AttrDict()
        for key, field in self.fields.items():
            value = field.load(obj)
            # If there is no value, don't include anything in the output dictionary
            if not is_valid(value):
                continue
            # If the value isn't a dict, or there's not flatten directive, add and continue
            if not isinstance(value, dict) or not getattr(field, "flatten", False):
                output[field.output_name] = value
                continue
            # Merge in flattened fields
            output.update(value)
        return output
    
    def convert_groupdict(self, obj):
        return obj

class ZipSchema(Schema):
    """
    This schema type allows fields that produce multiple values to be
    zipped together into records.
    """
    def __init__(self, *args, **kwargs):
        superclasses = type(self).mro()
        type_class = [s for s in superclasses if issubclass(s, ZipSchema)][-2]
        module_path = type_class.__module__.split(".")
        module = ".".join(module_path[:-1] + ["fields",])
        self.list_field = getattr(importlib.import_module(module), "List")
        super().__init__(*args, **kwargs)

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        new_fields = dict()
        for name, field in self.fields.items():
            list_field = self.list_field(field.__class__, field.data_key)
            list_field.output_name = field.output_name
            new_fields[name] = list_field
        self.fields = new_fields

    def deserialize(self, obj) -> "Dict":
        result = unzip_records(super().deserialize(obj))
        return result

    def load_model(self, obj):
        return [self.__model__(**o) for o in obj]


class AttrDict(dict):    
    def __getattr__(self,  name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)
    
    def __setattr__(self, name, value):
        self[name] = value
    
    @classmethod
    def convert(cls, obj):
        if isinstance(obj, dict):
            return cls((k, cls.convert(v)) for k, v in obj.items())
        elif isinstance(obj, list):
            return list(cls.convert(i) for i in obj)
        else:
            return obj

    def to_json(self, *args, **kwargs):
        return json.dumps(self, *args, default=date_encoder, **kwargs)


from yankee.base import fields
from .schema import PolymorphicSchema, Schema, RegexSchema, ZipSchema

try:
    import jsonpath_ng
    def JsonPath(string):
        try:
            return jsonpath_ng.parse(string)
        except jsonpath_ng.exceptions.JsonPathParserError:
            return jsonpath_ng.ext.parse(string)
except ImportError:
    pass




import re
import itertools
import datetime
import ujson as json
import dataclasses as dc

us_re_1 = re.compile(r"([A-Z]+)([A-Z][a-z])")
us_re_2 = re.compile(r"([a-z\d])([A-Z])")
us_re_3 = re.compile(r"([^\d_])(\d+)")


def underscore(word: str) -> str:
    """
    Modified version from inflection library
    that also underscores digits
    """
    word = us_re_1.sub(r"\1_\2", word)
    word = us_re_2.sub(r"\1_\2", word)
    word = us_re_3.sub(r"\1_\2", word)
    word = word.replace("-", "_")
    return word.lower()


camelize_re = re.compile(r"(?:^|_)(.)")


def camelize(string: str) -> str:
    """
    Optimized version of the camelize function in inflections
    """
    result = camelize_re.sub(lambda m: m.group(1).upper(), string)
    return result[0].lower() + result[1:]


def is_valid(obj):
    if obj is None:
        return False
    elif isinstance(obj, (int, float)):
        return True
    elif isinstance(obj, (dict, list, str)) and len(obj) == 0:
        return False
    elif dc.is_dataclass(obj):
        return bool(obj)
    return True

def inflect(string, style=None):
    try:
        if style is None:
            return string
        elif style == "json":
            return camelize(string)
        elif style == "python":
            return underscore(string)
    except Exception:
        return None

# Cleans whitespace from text data
whitespace_re = re.compile(r"\s+")

def clean_whitespace(s, preserve_newlines=False):
    if preserve_newlines:
        return "\n".join(whitespace_re.sub(" ", l).strip().strip(",").strip() for l in s.split("\n")).strip()
    else:
        return whitespace_re.sub(" ", s).strip().strip(",").strip()

strip_lines = lambda s: "\n".join(l.strip() for l in s.split("\n"))

def do_nothing(obj):
    return obj

def date_encoder(obj):
    if isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()
    raise ValueError(f"Object of type {type(obj)} is not serializable!")

class AttrDict(dict):    
    def __getattr__(self,  name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)
    
    def __setattr__(self, name, value):
        self[name] = value
    
    @classmethod
    def convert(cls, obj):
        if isinstance(obj, dict):
            return cls((k, cls.convert(v)) for k, v in obj.items())
        elif isinstance(obj, list):
            return list(cls.convert(i) for i in obj)
        else:
            return obj

    def to_json(self, *args, **kwargs):
        return json.dumps(self, *args, default=date_encoder, **kwargs)

def update_class(orig, update):
    for k in filter(lambda k: not k.startswith("_"), update.__dict__.keys()):
        setattr(orig, k, getattr(update, k))

def unzip_records(data):
    keys = list(data.keys())
    value_lists = [data[k] for k in keys]
    return list(AttrDict(zip(keys, o)) for o in itertools.zip_longest(*value_lists, fillvalue=None))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yankee',
 'yankee.base',
 'yankee.base.fixtures',
 'yankee.base.fixtures.nested',
 'yankee.base.fixtures.nested.model',
 'yankee.base.fixtures.nested.schema',
 'yankee.data',
 'yankee.html',
 'yankee.html.io',
 'yankee.html.schema',
 'yankee.io',
 'yankee.io.test',
 'yankee.json',
 'yankee.json.schema',
 'yankee.xml',
 'yankee.xml.io',
 'yankee.xml.schema']

package_data = \
{'': ['*']}

install_requires = \
['cssselect>=1.2.0,<2.0.0',
 'jsonpath-ng>=1.5.3,<2.0.0',
 'lxml>=4.8.0,<5.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'ujson>=5.7.0,<6.0.0']

setup_kwargs = {
    'name': 'yankee',
    'version': '0.1.39',
    'description': 'lightweight, simple, and fast declarative XML and JSON data extraction',
    'long_description': '# Yankee - Simple Declarative Data Extraction from XML and JSON\n\nThis is kind of like Marshmallow, but only does deserialization. What it lacks in reversibility, it makes up for in speed. Schemas are compiled in advance allowing\ndata extraction to occur very quickly.\n\n## Motivation\n\nI have another package called patent_client. I also do a lot with legal data, some of which is in XML, and some of which is in JSON. But there\'s a lot of it. And I mean *a lot*, so speed matters.\n\n## Quick Start\n\nThere are two main modules: `yankee.json.schema` and `yankee.xml.schema`. Those modules support defining class-style deserializers. Both start by subclassing a `Schema` class, and then defining attributes from the `fields` submodule.\n\n### JSON Deserializer Example\n\n```python\n    from yankee.json import Schema, fields\n\n    class JsonExample(Schema):\n        name = fields.String()\n        birthday = fields.Date("birthdate")\n        deep_data = fields.Int("something.0.many.levels.deep")\n\n    obj = {\n        "name": "Johnny Appleseed",\n        "birthdate": "2000-01-01",\n        "something": [\n            {"many": {\n                "levels": {\n                    "deep": 123\n                }\n            }}\n        ]\n    }\n\n    JsonExample().deserialize(obj)\n    # Returns\n    {\n        "name": "Johnny Appleseed",\n        "birthday": datetime.date(2000, 1, 1),\n        "deep_data": 123\n    }\n\n```\n\nFor JSON, the attributes are filled by pulling values off of the JSON object. If no\npath is provided, then the attribute name is used. Otherwise, a dotted string\ncan be used to pluck an item from the JSON object.\n\n### XML Deserializer Example\n\n```python\n    import lxml.etree as ET\n    from yankee.xml import Schema, fields\n\n    class XmlExample(Schema):\n        name = fields.String("./name")\n        birthday = fields.Date("./birthdate")\n        deep_data = fields.Int("./something/many/levels/deep")\n\n    obj = ET.fromstring(b"""\n    <xmlObject>\n        <name>Johnny Appleseed</name>\n        <birthdate>2000-01-01</birthdate>\n        <something>\n            <many>\n                <levels>\n                    <deep>123</deep>\n                </levels>\n            </many>\n        </something>\n    </xmlObject>\n    """.strip())\n\n    XmlExample().deserialize(obj)\n    # Returns\n    {\n        "name": "Johnny Appleseed",\n        "birthday": datetime.date(2000, 1, 1),\n        "deep_data": 123\n    }\n```\n\nFor XML, the attributes are filled using XPath expressions. If no path is provided,\nthen the entire object is passed to the field (no implicit paths). Any valid Xpath\nexpression can be used.\n\n',
    'author': 'Parker Hancock',
    'author_email': '633163+parkerhancock@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/parkerhancock/gelatin_extract',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections.abc import Sequence
from yankee.util import camelize
from yankee.base.accessor import do_nothing
import jsonpath_ng


def json_accessor(data_key, name, many, meta):
    # Handle JSONPath objects passed as data keys
    if isinstance(data_key, jsonpath_ng.JSONPath):
        def accessor_func(obj):
            result = [match.value for match in data_key.find(obj)]
            try:
                return result if many else result[0]
            except IndexError:
                return None
        return accessor_func
    # Handle no data key
    if data_key is False or (data_key is None and name is None):
        return do_nothing
    # Handle implicit data keys
    elif data_key is None and meta.infer_keys:
        data_key = camelize(name)
    key_segments = tuple(int(s) if s.isdigit() else s for s in data_key.split("."))
    def accessor_func(obj):
        result = obj
        for seg in key_segments:
            if isinstance(obj, Sequence):
                try:
                    result = result[int(seg)]
                except IndexError:
                    result = None
            else:
                result = result.get(seg, None)
        return result
    return accessor_func


import io

from ...io.iterparse import file_iterparse


def aps_record_to_xml(text):
    if isinstance(text, str):
        buf = io.StringIO(text)
    else:
        buf = io.TextIOWrapper(io.BytesIO(text))
    record_tag = buf.readline().strip()
    output = list()
    output.append(f"<{record_tag}>")
    section_tag = None
    item_tag = None
    for line in buf.readlines():
        key = line[:5].strip()
        value = line[5:].strip()
        if key and not value:  # New Section
            if item_tag:
                output.append(f"</{item_tag}>")
                item_tag = None
            if section_tag:
                output.append(f"</{section_tag}>")
            output.append(f"<{key}>")
            section_tag = key
        elif value and not key:
            if item_tag == "TBL":
                output.append(f"{value}\n")
            else:
                output.append(value)
        else:
            # New Item
            if item_tag:
                output.append(f"</{item_tag}>")
            output.append(f"<{key}>")
            item_tag = key
            if item_tag == "TBL":
                output.append(f"{value}\n")
            else:
                output.append(value)
    if item_tag:
        output.append(f"</{item_tag}>")
    if section_tag:
        output.append(f"</{section_tag}>")
    output.append(f"</{record_tag}>")
    return "".join(output).encode()


def aps_iterator(file_obj: io.RawIOBase, record_tag=None):
    for record in file_iterparse(file_obj, start=record_tag.encode()):
        yield aps_record_to_xml(record)



import lxml.etree as ET
from yankee.base.accessor import do_nothing

def xml_accessor(data_key, name, many, meta):
    if isinstance(data_key, ET.XPath):
        def accessor_func(obj):
            if obj is None:
                return None
            result = data_key(obj)
            return result if many else result[0]
        return accessor_func
    
    if not data_key:
        return do_nothing

    namespaces = getattr(meta, "namespaces", None)
    if many:
        xpath = ET.XPath(data_key, namespaces=namespaces)
    else:
        xpath = ET.XPath(f"({data_key})[1]", namespaces=namespaces)
    
    def accessor_func(obj):
        if obj is None:
            return None
        result = xpath(obj)
        try:
            return result if many else result[0]
        except IndexError:
            return None
    return accessor_func


import lxml.etree as ET
from yankee.base import schema
from yankee.base.deserializer import Deserializer

from .mixin import XmlMixin

class Deserializer(XmlMixin, Deserializer):
    pass

class Schema(XmlMixin, schema.Schema):
    def load(self, obj):
        if isinstance(obj, (ET._Element, ET._ElementTree)):
            return super().load(obj)
        elif isinstance(obj, str):
            return super().load(ET.fromstring(obj.encode()))
        elif isinstance(obj, bytes):
            return super().load(ET.fromstring(obj))


class PolymorphicSchema(XmlMixin, schema.PolymorphicSchema):
    pass

class RegexSchema(XmlMixin, schema.RegexSchema):
    pass

class ZipSchema(XmlMixin, schema.ZipSchema):
    pass